@@ -1,7 +1,9 @@
+import json
+import math
 from copy import deepcopy
 from functools import partial
 from types import FunctionType
-from typing import Type, Tuple, Sequence, Dict, Callable, Any, Optional
+from typing import Type, Tuple, Sequence, Dict, Callable, Any, Optional, List
 
 from django.apps import apps
 from django.conf.urls import url
@@ -425,26 +427,23 @@ class JsonApiModelViewBuilder:
         return self._build(url_resource_name=url_resource_name, urls_prefix=urls_prefix)
 
 
-class JsonApiActionViewBuilder:
-    def __init__(self, action_name: str = None,
+class JsonApiResourceViewBuilder:
+    def __init__(self,
+                 action_name: str = None,
+                 unique_identifier: Optional[str] = 'id',
                  allowed_methods=json_api_spec_http_methods.HTTP_ACTIONS,
                  permission_classes: Sequence[Type[BasePermission]] = None,
                  authentication_classes: Sequence[Type[BaseAuthentication]] = None):
-        self.__validate_http_methods(allowed_methods)
         self._allowed_methods = [*allowed_methods]
         self._resource_name = action_name
+        self._unique_identifier = unique_identifier
         self._permission_classes = permission_classes or []
         self._authentication_classes = authentication_classes or []
         self._on_create_callback = None
         self._on_update_callback = None
         self._on_delete_callback = None
-
-    @staticmethod
-    def __validate_http_methods(limit_to_http_methods: Sequence[str] = json_api_spec_http_methods.HTTP_ALL):
-        if any(map(lambda method: method not in json_api_spec_http_methods.HTTP_ACTIONS, limit_to_http_methods)):
-            raise Exception(
-                f'Cannot limit actions fields to HTTP Method of types: '
-                f'{list(filter(lambda method: method not in json_api_spec_http_methods.HTTP_ACTIONS, limit_to_http_methods))}')
+        self._on_list_callback = None
+        self._on_get_callback = None
 
     def __warn_if_method_not_available(self, method: str):
         if method not in self._allowed_methods:
@@ -452,35 +451,86 @@ class JsonApiActionViewBuilder:
                 f'You\'ve set a lifecycle callback for resource {self._resource_name}, '
                 f'which doesn\'t allow it\'s respective HTTP method through `allowed_methods`.')
 
-    def on_create(self, create_callback: Callable[..., Response] = None) -> 'JsonApiActionViewBuilder':
+    def on_create(self,
+                  create_callback: Callable[[Request], Tuple[Dict, str, int]] = None) -> 'JsonApiResourceViewBuilder':
         self._on_create_callback = create_callback
         self.__warn_if_method_not_available(json_api_spec_http_methods.HTTP_POST)
         return self
 
-    def on_update(self, update_callback: Callable[..., Response] = None) -> 'JsonApiActionViewBuilder':
+    def on_update(self, update_callback: Callable[[Request], Tuple[Dict, int]] = None) -> 'JsonApiResourceViewBuilder':
         self._on_update_callback = update_callback
         self.__warn_if_method_not_available(json_api_spec_http_methods.HTTP_PATCH)
         return self
 
-    def on_delete(self, delete_callback: Callable[..., Response] = None) -> 'JsonApiActionViewBuilder':
+    def on_delete(self, delete_callback: Callable[[Request], Tuple[int]] = None) -> 'JsonApiResourceViewBuilder':
         self._on_delete_callback = delete_callback
         self.__warn_if_method_not_available(json_api_spec_http_methods.HTTP_DELETE)
         return self
 
+    def on_list(self, list_callback: Callable[[Request], Tuple[List, int, int]] = None) -> 'JsonApiResourceViewBuilder':
+        self._on_list_callback = list_callback
+        self.__warn_if_method_not_available(json_api_spec_http_methods.HTTP_GET)
+        return self
+
+    def on_get(self, get_callback: Callable[[Request], Tuple[Dict, int]] = None) -> 'JsonApiResourceViewBuilder':
+        self._on_get_callback = get_callback
+        self.__warn_if_method_not_available(json_api_spec_http_methods.HTTP_GET)
+        return self
+
     def _build(self, url_resource_name: str = '', urls_prefix: str = '', urls_suffix: str = '') -> Sequence[partial]:
         def destroy(view, request, *args, **kwargs):
+            identifier = kwargs.get(self._unique_identifier, None)
+
             if self._on_delete_callback is not None:
-                self._on_delete_callback(request, *args, **kwargs)
+                status = self._on_delete_callback(request, identifier, *args, **kwargs)
+                return Response(data={}, status=status)
 
         def update(view, request, *args, **kwargs):
+            data = json.loads(request.body).get('data', {})
+            identifier = kwargs.get(self._unique_identifier, None)
+
             if self._on_update_callback is not None:
-                self._on_update_callback(request, *args, **kwargs)
+                data, status = self._on_update_callback(request, identifier, data, *args, **kwargs)
+                return Response(data={'id': identifier, 'type': self._resource_name, 'attributes': data}, status=status)
 
         def create(view, request, *args, **kwargs):
+            data = json.loads(request.body).get('data', {})
             if self._on_create_callback is not None:
-                self._on_create_callback(request, *args, **kwargs)
+                data, identifier, status = self._on_create_callback(request, data, *args, **kwargs)
+                return Response(
+                    data={'id': identifier, 'type': self._resource_name, 'attributes': data}, status=status)
 
-        view_set = type(f'{self._resource_name}JSONApiActionViewSet', (ViewSet,), {
+        def _list(view, request, *args, **kwargs):
+            params = request.query_params
+            page = params.get('page_number', 1)
+            if self._on_list_callback is not None:
+                data, count, status = self._on_list_callback(request, page, *args, **kwargs)
+                pages = math.ceil(count / 50)
+                return Response(data={'links': {
+                    "first": f"/api/{self._resource_name}?page_number=1",
+                    "last": f"/api/{self._resource_name}?page_number={pages}",
+                    "next": None if page == pages or pages <= 1 else f"/api/{self._resource_name}?page_number={page + 1}",
+                    "previous": None if page <= 1 else f"/api/{self._resource_name}?page_number={page - 1}",
+                }, 'data': [
+                    {'id': item.get(self._unique_identifier, None), 'type': self._resource_name,
+                     'attributes': item} for
+                    item in data], 'meta': {
+                    'pagination': {
+                        "page": page,
+                        "pages": pages,
+                        "count": count
+                    }
+                }
+                }, status=status)
+
+        def get(view, request, *args, **kwargs):
+            identifier = kwargs.get(self._unique_identifier, None)
+
+            if self._on_get_callback is not None:
+                data, status = self._on_get_callback(request, identifier, *args, **kwargs)
+                return Response(data={'id': identifier, 'type': self._resource_name, 'attributes': data}, status=status)
+
+        patch_view_set = type(f'{self._resource_name}ChangeJSONApiActionViewSet', (ViewSet,), {
             'renderer_classes': (JSONRenderer, BrowsableAPIRenderer),
             'parser_classes': (JSONParser, FormParser, MultiPartParser),
             'metadata_class': JSONAPIMetadata,
@@ -491,9 +541,24 @@ class JsonApiActionViewBuilder:
             'http_method_names': list(map(lambda method: method.lower(), self._allowed_methods)) + ['head', 'options'],
             'permission_classes': self._permission_classes,
             'authentication_classes': self._authentication_classes,
-            'create': create,
             'update': update,
-            'destroy': destroy
+            'destroy': destroy,
+            'get': get
+        })
+
+        get_view_set = type(f'{self._resource_name}RetrieveJSONApiActionViewSet', (ViewSet,), {
+            'renderer_classes': (JSONRenderer, BrowsableAPIRenderer),
+            'parser_classes': (JSONParser, FormParser, MultiPartParser),
+            'metadata_class': JSONAPIMetadata,
+            'pagination_class': LimitedJsonApiPageNumberPagination,
+            'filter_backends': (
+                QueryParameterValidationFilter, OrderingFilter, JsonApiSearchFilter),
+            'resource_name': None,
+            'http_method_names': list(map(lambda method: method.lower(), self._allowed_methods)) + ['head', 'options'],
+            'permission_classes': self._permission_classes,
+            'authentication_classes': self._authentication_classes,
+            'list': _list,
+            'create': create,
         })
 
         urls = []
@@ -504,11 +569,15 @@ class JsonApiActionViewBuilder:
         if len(url_resource_name) == 0:
             url_resource_name = self._resource_name
 
+
         urls.extend([
             url(rf'^{urls_prefix}{url_resource_name}{urls_suffix}$',
-                view_set.as_view({'post': 'create', 'patch': 'update', 'delete': 'destroy'},
-                                 name=f'list_{self._resource_name}'),
-                name=f'{self._resource_name}-action'),
+                get_view_set.as_view({'get': 'list', 'post': 'create'}, name=f'list_{self._resource_name}'),
+                name=f'list-{self._resource_name}-action'),
+            url(rf'^{urls_prefix}{url_resource_name}{urls_suffix}/(?P<{self._unique_identifier}>[^/.]+)/$',
+                patch_view_set.as_view({'get': 'get', 'patch': 'update', 'delete': 'destroy'},
+                                       name=f'patch_destroy_{self._resource_name}'),
+                name=f'patch-delete-{self._resource_name}-action')
         ])
         return urls
 
@@ -516,18 +585,21 @@ class JsonApiActionViewBuilder:
         return self._build(url_resource_name=url_resource_name, urls_prefix=urls_prefix, urls_suffix=urls_suffix)
 
 
-def json_api_action_view(action_name: str,
-                         method: str = json_api_spec_http_methods.HTTP_POST,
-                         permission_classes: Sequence[Type[BasePermission]] = None,
-                         authentication_classes: Sequence[Type[BaseAuthentication]] = None,
-                         urls_prefix: str = '',
-                         urls_suffix: str = '', ) -> FunctionType:
-    def decorator(func):
-        builder = JsonApiActionViewBuilder(action_name=action_name,
-                                           allowed_methods=[method],
-                                           permission_classes=permission_classes,
-                                           authentication_classes=authentication_classes)
-        if method == json_api_spec_http_methods.HTTP_POST:
+def json_api_view(resource_name: str,
+                  method: str = json_api_spec_http_methods.HTTP_GET,
+                  permission_classes: Optional[Sequence[Type[BasePermission]]] = None,
+                  authentication_classes: Optional[Sequence[Type[BaseAuthentication]]] = None,
+                  urls_prefix: str = '',
+                  urls_suffix: str = '', ) -> FunctionType:
+    def decorator(func: Callable[[Request], Tuple[Dict, int]]):
+        builder = JsonApiResourceViewBuilder(action_name=resource_name,
+                                             allowed_methods=[method],
+                                             permission_classes=permission_classes,
+                                             authentication_classes=authentication_classes)
+        if method == json_api_spec_http_methods.HTTP_GET:
+            return builder.on_list(list_callback=func) \
+                .get_urls(urls_prefix=urls_prefix, urls_suffix=urls_suffix)
+        elif method == json_api_spec_http_methods.HTTP_POST:
             return builder.on_create(create_callback=func) \
                 .get_urls(urls_prefix=urls_prefix, urls_suffix=urls_suffix)
         elif method == json_api_spec_http_methods.HTTP_DELETE:
